/*
 * ans_formula - Analyze-and-store Formula.  Process a numeric formula, tokenize
 *		 and store it in stmt.
 */

#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include "hptsb.h"
#include "hptoken.h"
#include "decl.h"

/*
 * ans_formula will tokenize and store numeric expressions, including variables and functions.
 */

uint8_t *
ans_formula (uint8_t **lp, uint32_t *sttype, uint8_t *stmt, uint32_t *stlen, uint8_t *varname, uint8_t *vartype, uint8_t *varnum)
{
	uint8_t		*newlp;
	uint32_t	newsttype;
	float		numop;
	uint32_t	HPfloat = 0;
	uint32_t	parenlevel = 0;
	uint32_t	operatorlast = 1;
	uint32_t	initialmatch = 0;
	uint32_t	stored = 0;
	uint8_t		*save_lp = *lp;
	uint8_t		*save_stmt = stmt;
	uint32_t	save_stlen = *stlen;
	uint32_t	save_sttype = *sttype;
	uint8_t		save_varname = *varname;
	uint8_t		save_vartype = *vartype;
	uint8_t		save_varnum = *varnum;
	uint8_t		*nstmt;
	int32_t        	stringlen;
	uint8_t         litstring[512];

	/* The following do not need pushes onto the save stack */


	dprintf (DEBUG_TOKEN, "IN ans_formula: *sttype = %d\n", *sttype);
	dprintf (DEBUG_TOKEN, "*********** ADDRESS OF newlp = %x\n", &newlp);
	dprintf (DEBUG_TOKEN, "ANS: INITIAL initialmatch = 0\n");
	
	if (**lp == 0)
		return (stmt);		/* If nothing to do, just exit without storing sttype */

	while (**lp != 0) {
		/* Check for (, ), [ and ]. */
		if (**lp == '(' || **lp == '[') {
			parenlevel++;
			dprintf (DEBUG_TOKEN, "((((((((((((((((((((((((((((((((((( NEW PAREN LEVEL=%d\n", parenlevel);
			dprintf (DEBUG_TOKEN, "ANS: initialmatch=0\n");
			initialmatch = 0;
		} else if (**lp == ']' || **lp == ')') {
				dprintf (DEBUG_TOKEN, "))))))))))))))))))))))))))))))))))) NEW PAREN LEVEL=%d\n", parenlevel);
			if (parenlevel == 0) {
				/* Yes, if we already have an operator, write that first, then save this one. */
				dprintf (DEBUG_TOKEN, "ans_formula: 3-Storing sttype=%o, varname = %d  vartype=%o\n",
						*sttype, *varname, *vartype);
				if (*sttype != 0) {
					dprintf (DEBUG_TOKEN, "ans_formula: 4-Storing sttype=%o, varname = %d  vartype=%o\n",
							*sttype, *varname, *vartype);
					*stlen += 1;
					/* We'll only have an operator, so no need for the operand values in stmt */
					*stmt++ = ITOWH (*sttype << 9);
					*stmt++ = ITOWL (*sttype << 9);
					stored++;
					*sttype = 0;
					*varname = 0;
					*vartype = 0;
					*varnum = 99;
				}
				dprintf (DEBUG_TOKEN, "ans_formula: RETURNING with **lp=%c\n", **lp);
				return (stmt);
			}
			parenlevel--;
		}
		dprintf (DEBUG_TOKEN, "ans_formula: Checking pos %x ('%c')\n", *lp, **lp);
		/* Is the next match an operator? */
		newlp = match_token (*lp, &newsttype, Bop_ACCESS, Bop_ACCESS_count, 1);
		dprintf (DEBUG_TOKEN, "ans_formula: lp=%x   newlp=%x  sttype=%o newsttype=%o\n",
				*lp, newlp, *sttype, newsttype);
		dprintf (DEBUG_TOKEN, "ans_formula: CHECK: **lp=%c  **lp++=%c *sttype=%o\n",
				**lp, *get_nth_character(*lp, 1), *sttype);

		if (*lp != newlp && ( (newsttype != 014 && newsttype != 015) || !isdigit(*get_nth_character(*lp, 1)))) {
		dprintf (DEBUG_TOKEN, "ANS_FORMULA: Passed test 1\n");

			if (newsttype == 017)
				newsttype = 030; /* Change "assignment =" to "relational =" */
			else if ( newsttype == 016 ||
					  (!initialmatch && (newsttype && newsttype != 010 && newsttype != 011
						   && newsttype != 013 && newsttype != 012 && newsttype != 027
						   && newsttype != 025 && newsttype != 026 && newsttype != 034
						   && newsttype != 035 && newsttype != 042
						   && newsttype != 041 && newsttype != 030 && newsttype != 032
						   && newsttype != 033 && newsttype != 022 && newsttype != 023
						   && newsttype != 020 && newsttype != 021 && newsttype != 024
						   && newsttype != 036 && newsttype != 037 && newsttype != 040))) {

				dprintf (DEBUG_TOKEN, "ERROR: Can't start a numeric formula with '%c', tokencode=%o\n",
						**lp, newsttype);
				if (*sttype != 0) {
					*stlen += 1;
					*stmt++ = ITOWH (*sttype << 9);
					*stmt++ = ITOWL (*sttype << 9);
					*sttype = 0;
				}
				return (stmt);
			}
			dprintf (DEBUG_TOKEN, "ans_formula: New operator found, sttype=%d.\n", *sttype);
			/* Yes, if we already have an operator, write that first, then save this one. */
			if (*sttype != 0) {
			dprintf (DEBUG_TOKEN, "ans_formula: 1-Storing sttype=%o, varname = %d  vartype=%o\n",
						*sttype, *varname, *vartype);
				*stlen += 1;
				/* We'll only have an operator, so no need for the operand values in stmt */
				*stmt++ = ITOWH (*sttype << 9);
				*stmt++ = ITOWL (*sttype << 9);
				stored++;
				*sttype = 0;
				*varname = 0;
				*vartype = 0;
				*varnum = 99;
			}
			*sttype = newsttype;
			*lp = get_next_character(newlp);
			dprintf (DEBUG_TOKEN, "ANS2: initialmatch=0\n");
			initialmatch = 0;
		} else {
			dprintf (DEBUG_TOKEN, "ans_formula: Checking for a function... operatorlast=%d  **lp=%c\n",
					operatorlast, **lp);
			newlp = match_token (*lp, &newsttype, Bfunc, Bfunc_count, 1);
			if (operatorlast && *lp != newlp) {
uint8_t *x_lp;
				dprintf (DEBUG_TOKEN, "ans_formula: FUNCTION found: %o\n", newsttype);
				*lp = newlp;
				*varname = newsttype;
				*vartype = 017;
				*stlen += 1;
				*stmt++ = ITOWH (0x8000 | (*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
				*stmt++ = ITOWL (0x8000 | (*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
				stored++;
				dprintf (DEBUG_TOKEN, "ans_formula: STORED sttype=%o, varname=%o  vartype=%o\n",
						*sttype, *varname, *vartype);
				*sttype = 0;
				*varname = 0;
				*vartype = 0;
				*varnum = 99;
				dprintf (DEBUG_TOKEN, "+++++++++++++++++++++++++++ Need to mandate ( or [ after function\n");

				if (**lp != '(' && **lp != '[') {
					dprintf (DEBUG_TOKEN, "ans_formula: Missing ( or [ function name\n");
					return (save_stmt);
				}
				*sttype = 013;
				*lp = get_next_character (*lp + 1);

				if (newsttype == 015 || newsttype == 025 || newsttype == 026) {
								/* 015 == LEN(), 025 == NUM(), 026 == POS() */

					newlp = get_source_string (*lp, varname, vartype, litstring, &stringlen);
					*sttype = 013;
					if (stringlen != 9999) {
						*stlen += 1;
						*stmt++ = ITOWH ((*sttype << 9));
						*stmt++ = ITOWL ((*sttype << 9));
						*sttype = 1;  
						*stlen += 1;
						*stmt++ = ITOWH (*sttype << 9 | (stringlen & 0377));
						*stmt++ = ITOWL (*sttype << 9 | (stringlen & 0377));
						memcpy (stmt, litstring, stringlen);
						*stlen += (stringlen + 1) / 2;
						stmt += stringlen + (stringlen % 2);
						dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
						*sttype = 0;
						*lp = newlp;
					} else if (newlp != *lp) {
						*lp = newlp;
						*stlen += 1;
						*stmt++ = ITOWH ((*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
						*stmt++ = ITOWL ((*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
						dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n",
						*varname, *vartype, stringlen);
						*sttype = 0;
						nstmt = get_subscript (lp, sttype, stmt, stlen);
						if (nstmt == NULL) {
							/*eprintf ("ERROR: Invalid subscript for string variable.\n");*/
							/*exit (80);*/
							return (save_stmt);
						}
						stmt = nstmt;
					} else {
						return (save_stmt);  /* No string expression found, so reset and exit */
					}

					if (newsttype == 026) { /* Get second string argument for POS() */
						*lp = get_next_character (*lp);
						if (**lp != ',') {
							return (save_stmt);  /* Missing comma for POS() parameters, so reset and exit */
						}
						*sttype = 2;  
						*lp = get_next_character (*lp + 1);

						newlp = get_source_string (*lp, varname, vartype, litstring, &stringlen);
/*
eprintf("WHAMMMMMMMMMMMM     GOT LEN$()\n");
*/
						dprintf (DEBUG_TOKEN, "ans_formula: get_source_string: RETURNED:  *lp=%x  newlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d\n",
								*lp, newlp, *varname, *vartype, litstring, stringlen);
						if (stringlen != 9999) {
							*stlen += 1;
							*stmt++ = ITOWH ((*sttype << 9));
							*stmt++ = ITOWL ((*sttype << 9));
							*sttype = 1;  
							*stlen += 1;
							*stmt++ = ITOWH (*sttype << 9 | (stringlen & 0377));
							*stmt++ = ITOWL (*sttype << 9 | (stringlen & 0377));
							memcpy (stmt, litstring, stringlen);
							*stlen += (stringlen + 1) / 2;
							stmt += stringlen + (stringlen % 2);
							dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
							*lp = newlp;
						} else if (newlp != *lp) {
							*lp = newlp;
							*stlen += 1;
							*stmt++ = ITOWH ((*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
							*stmt++ = ITOWL ((*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
							dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n",
							*varname, *vartype, stringlen);
							*sttype = 0;
							nstmt = get_subscript (lp, sttype, stmt, stlen);
							if (nstmt == NULL) {
								/*eprintf ("ERROR: Invalid subscript for string variable.\n");*/
								/*exit (80);*/
								return (save_stmt);
							}
							stmt = nstmt;
						} else {
							return (save_stmt);  /* No string expression found, so reset and exit */
						}
					}

/* END LEN/NUM/POS processing */
				} else { /* Not a special numeric function taking a string argument. */
					nstmt = ans_formula (lp, sttype, stmt, stlen, varname, vartype, varnum);


				if (nstmt == NULL) {
					dprintf (DEBUG_TOKEN, "%%%%%%%%%%%% ans_formula bailing, recall to ansformula null.\n");
					*lp = save_lp;
					*sttype = save_sttype;
					stmt = save_stmt;
					*stlen = save_stlen;
					*varname = save_varname;
					*vartype = save_vartype;
					*varnum = save_varnum;
					return (save_stmt);
				}
					stmt = nstmt;
				}

					if (**lp != ')' && **lp != ']') {
					    	dprintf (DEBUG_TOKEN, "ans_formula: Missing ) or ] after function name\n");
					    	return (save_stmt);
					}

					*sttype = 010;
					*stlen += 1;
					*stmt++ = ITOWH (*sttype << 9);
					*stmt++ = ITOWL (*sttype << 9);
					*lp = get_next_character (*lp + 1);
					dprintf (DEBUG_TOKEN, "ans_formula: FINISHED with function processing--old lp=%x  new lp=%x.\n",
							x_lp, lp);
					*sttype = 0;
					dprintf (DEBUG_TOKEN, "ANS4: initialmatch=1\n");
					initialmatch = 1;



			/*
			 * No, check for a numeric operand.
			 * If found, save it and write the current opair.
			 *
			 * If not an operand either, exit.
			 */
			} else if (operatorlast && (newlp = get_number (*lp, GN_FLOAT, &numop)) != NULL) {

				/*newlp = get_number (*lp, GN_FLOAT, &numop);*/
				HPfloat = IEEEtoHPf (numop);
				*varname = newsttype;
				*vartype = 017;
				*stlen += 1;
				*stmt++ = ITOWH (0x8000 | (*sttype << 9));
				*stmt++ = ITOWL (0x8000 | (*sttype << 9));
				stored++;
				*stlen += 1;
				*stmt++ = ITOWH (HPfloat >> 16);	/* Sign + High Mantissa */
				*stmt++ = ITOWL (HPfloat >> 16);
				stored++;
				*stlen += 1;
				*stmt++ = ITOWH (HPfloat & 0xFFFF);	/* Low Matissa, Exp + Sign */
				*stmt++ = ITOWL (HPfloat & 0xFFFF);
				stored++;
				dprintf (DEBUG_TOKEN, "ans_formula: STORE an HP Floating Point number, sttype=%o\n",
						*sttype);
				*sttype = 0;
				*varname = 0;
				*vartype = 0;
				*varnum = 99;
				*lp = newlp;
				dprintf (DEBUG_TOKEN, "ANS3: initialmatch=1\n");
				initialmatch = 1;
			} else if (isalpha (**lp) && operatorlast) {   /* Nope, check for a variable now */
				/*** SAME AS PREVIOUS VARIABLE CHECKING FROM HERE ON (mostly, except for string stuff) !!! ***/
				newlp = *lp;
				*vartype = 0;
				*varname = 0;
				if (**lp == 'F' && *(get_next_character(*lp + 1)) == 'N') {
					*vartype=017;
					*lp = get_next_character (*lp + 1);
					*lp = get_next_character (*lp + 1);
				}
				if (isalpha(**lp)) {
					dprintf (DEBUG_TOKEN, "ans_formula: GOT A VARIABLE LETTER: %c  -- sttype=%o\n",
							**lp, *sttype);
					if (!*vartype)
						*vartype = 4;
					*varname = **lp - '@';
					*lp = get_next_character (*lp + 1);
				}
				dprintf(DEBUG_TOKEN, "CHECKING VARIABLE FOR NUMERIC PART: *vartype=%o  isdigit(**lp)=%d  **lp=%c\n", *vartype, isdigit(**lp), **lp);
				if (/*  THIS IS LEFTOVER FOR FNx CHECKING */ *vartype != 017 && isdigit (**lp)) {
					dprintf (DEBUG_TOKEN, "setting for X1\n");
					*varnum = **lp - '0';
					*vartype = 5 + *varnum;
					*lp = get_next_character (*lp + 1);
				}

				if (*sttype == 0 && stored) {/* Can't store a variable without an operator */
					dprintf (DEBUG_TOKEN, "ans_formula: Not storing variable: sttype == 0\n");
					*lp = newlp;
					return (stmt);
				}

				dprintf (DEBUG_TOKEN, "ans_formula: VARIABLE found: vartype=%o  varname=%c(%o)  varnum=%d\n",
						*vartype, *varname, *varname, *varnum);
				dprintf (DEBUG_TOKEN, "STORING Variable found: %c%c%c   sttype=0%o  stmt=0x%x\n", *varname + '@',
                                		*varnum == 99 ? 0 : *varnum + '0',
                                		(*vartype == 5 || *vartype == 6 ) ? '$' : 0,
						*sttype, stmt);
				*stlen += 1;
				*stmt++ = ITOWH ((*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
				*stmt++ = ITOWL ((*sttype << 9) | ((*varname & 037) << 4) | (*vartype & 017));
				dprintf (DEBUG_TOKEN, "STORED Variable found: %c%c%c   sttype=0%o  stmt=0x%x\n", *varname,
                                		*varnum == 99 ? 0 : *varnum + '0',
                                		(*vartype == 5 || *vartype == 6 ) ? '$' : 0,
						*sttype, stmt);
				stored++;
				*sttype = 0;
				*varname = 0;
			newsttype = *vartype;
				*vartype = 0;
				*varnum = 99;
				dprintf (DEBUG_TOKEN, "ANS5: initialmatch=1\n");
				initialmatch = 1;

				*lp = get_next_character (*lp);
				if (**lp == '(' || **lp == '[') {

				    *sttype = 0;

/*eprintf ("XYZ: calling get_subscript, varname=%d vartype=%o newsttype=%o lp=%s\n", *varname, *vartype, newsttype, *lp);*/
				    if (newsttype == 017)
					*stlen |= 0x8000;
				    nstmt = get_subscript (lp, sttype, stmt, stlen);
				*stlen &= 0x7fff;

				    if (nstmt == NULL) {
					*lp = save_lp;
					*sttype = save_sttype;
					stmt = save_stmt;
					*stlen = save_stlen;
					*varname = save_varname;
					*vartype = save_vartype;
					*varnum = save_varnum;
					/*return (save_stmt);*/
				    } else {

					operatorlast = 2;
					stmt = nstmt;
					dprintf (DEBUG_TOKEN, "ANS4: initialmatch=1\n");
					initialmatch = 1;
				    }
				}
			} else {
				dprintf (DEBUG_TOKEN, "ans_formula: EXITING: Invalid expression at character: %c.\n", **lp);
                                if (*sttype != 0 && stored) { /* Leftover sttype operator to store before exiting? */
					dprintf (DEBUG_TOKEN, "ans_formula: 2-Storing sttype=%o, varname = %d  vartype=%o\n",
							*sttype, *varname, *vartype);
                                        *stlen += 1;
                                        /* We'll only have an operator, so no need for the operand values in stmt */
                                        *stmt++ = ITOWH (*sttype << 9);
                                        *stmt++ = ITOWL (*sttype << 9);   
					stored++;
                                        *sttype = 0;
                                        *varname = 0;
                                        *vartype = 0;
                                        *varnum = 99;
                                }
				return (stmt);
			}
		}
		if ((*sttype >= 012 && *sttype <= 042 ) || operatorlast == 2) /* Can a variable follow the last operator? */
			operatorlast = 1;
		else
			operatorlast = 0;
	}
	if (*sttype != 0) {
		*stlen += 1;
		*stmt++ = ITOWH (*sttype << 9);
		*stmt++ = ITOWL (*sttype << 9);
		stored++;
		*sttype = 0;
		*varname = 0;
		*vartype = 0;
		*varnum = 99;
	}
	*stlen += 1;
	*stmt++ = 0;
	*stmt++ = 0;
	stored++;
	return (stmt);
}

uint8_t *
get_number (uint8_t *lp, uint32_t type, void *val)
{
	float		numsign = 1;
	int32_t		expsign = 0;
	int32_t		exp = 0;
	float		fexp = 0.0;
	float		numop = 0.0;
	uint32_t	numfound = 0;
/* GN_FLOAT, GN_INT, GN_LINENO */

	if ((type & GN_LINENO) && !isdigit(*lp)) {
		dprintf (DEBUG_TOKEN, "get_number: Exiting -- first char not a digit: lp=%d '%c'\n", *lp, *lp);
		return NULL;
	}
	if (isdigit(*lp) || *lp == '-' || *lp == '+' || *lp == '.') {
		if ((type & GN_FLOAT) && *lp == '+') {
			dprintf (DEBUG_TOKEN, "get_number: EXPLICIT POSITIVE number from '%c'\n", *lp);
			lp = get_next_character (lp + 1);
		} else if ((type & GN_FLOAT) && *lp == '-') {
			dprintf (DEBUG_TOKEN, "get_number: EXPLICIT NEGATIVE number\n");
			lp = get_next_character (lp + 1);
			numsign = -1;
		} else {
			dprintf (DEBUG_TOKEN, "get_number: IMPLICIT POSITIVE number\n");
		}
		if (!isdigit(*lp) && *lp != '.') {
			dprintf (DEBUG_TOKEN, "ERROR: Invalid numeric literal '%c'\n", *lp);
			return (NULL);
		}
		numfound++;
		/* Extract a numeric operand */
		while (*lp != 0 && isdigit (*lp)) {
			numop = numop * 10 + (*lp - '0');
			lp = get_next_character (lp + 1);
		}
		if ((type & GN_FLOAT) && *lp == '.') {
			lp = get_next_character (lp + 1);
			expsign = 0;
			fexp = 0;
			while (*lp != 0 && isdigit(*lp)) {
				dprintf (DEBUG_TOKEN, "FRACTION: IN  *lp=%c  expsign=%d fexp=%G\n", *lp, expsign, fexp);
				expsign++;
				fexp = fexp + (*lp - '0') * powf(10, -1.0 * expsign);
				lp = get_next_character (lp + 1);
				dprintf (DEBUG_TOKEN, "FRACTION: OUT  *lp=%c  expsign=%d fexp=%G\n", *lp, expsign, fexp);
			}
			dprintf (DEBUG_TOKEN, "FRACOMPUTE: IN  float=%G\n", numop);
			numop = numop + fexp;
			dprintf (DEBUG_TOKEN, "FRACOMPUTE: OUT  float=%G\n", numop);
		}

		if ((type & GN_FLOAT) && toupper(*lp) == 'E') {
			expsign = 1;
			exp = 0;
			lp = get_next_character (lp + 1);
			if (*lp == '+')
				lp = get_next_character (lp + 1);
			else if (*lp == '-') {
				expsign = -1;
				lp = get_next_character (lp + 1);
			}
			while (*lp != 0 && isdigit (*lp)) {
				exp = exp * 10 + (*lp - '0');
				lp = get_next_character (lp + 1);
			}
			if (exp > 38) {
				if (exp > 99) {
					dprintf (DEBUG_TOKEN, "ILLEGAL EXPONENT\n");
					return (NULL);
				}
				/* Under/Overflow possible */
				dprintf (DEBUG_TOKEN, "WARNING: Large exponent - overflow/underflow possible.\n");
			}
			dprintf (DEBUG_TOKEN, "EXPONENTIAL: IN  numop=%g  exp=%d  expsign=%d\n", numop, exp, expsign);
			numop = numop * powf (10, exp * expsign);
			dprintf (DEBUG_TOKEN, "EXPONENTIAL: OUT  numop=%g  exp=%d  expsign=%d\n", numop, exp, expsign);
		}
		numop = numop * numsign;
		dprintf (DEBUG_TOKEN, "get_number: NUMERIC OPERAND FOUND: numop=%g\n", numop);
	}
	if (!numfound)
		return (NULL);
	if (type & GN_FLOAT)
		*(float *)(val) = numop;
	else if (type & GN_INT)
		*(int32_t *)(val) = (int32_t) numop;
	else {
		if (numop > 9999) {
			dprintf (DEBUG_TOKEN, "get_number: Line number > 9999 found: %G\n", numop);
			return (NULL);
		}
		if (numop < 1) {
			dprintf (DEBUG_TOKEN, "get_number: Line number < 1 found: %G\n", numop);
			return (NULL);
		}
		*(int32_t *)(val) = (int32_t) numop;
	}
	return (lp);
}
